/************************************************************************************************
 * Copyright  (c) 2011, 2012, 2013 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

/*** Floating-Point Adder ***/
/***	
Author: Edin Kadric
	IC Group, University of Pennsylvania

Notes: Code based on the architecture described in 
	"Accurate Parallel Floating-Point Accumulation", in Arith 2013

// Inputs:
//	IEEE754 number A
//	IEEE754 number B
//	Boolean Data present signal
// Outputs:
//	IEEE754 number S (sum)
//	Boolean Data present signal
//
// Note: The 7 cycles are called e, e2, 1, 2, 3, 34, 4

***/

package FPA;

import Types::*;
import MSOsearch::*;

interface FPAIfc;
	method ActionValue#(Sonly) do_sum(IEEE754 a_in, IEEE754 b_in, Bool datapi_in);
endinterface

(*synthesize*)
module mkFPA(FPAIfc);

	///////////////
	// Variables //
	///////////////

	Reg#(IEEE754) srs <- mkRegU; //sum
	Reg#(Bool) datapo <- mkReg(False); // data present
	Sonly s = unpack(pack(tuple2(srs, datapo))); // Stores the S and datapo signal

       	Reg#(Mantissa54) a1 <- mkRegU;
       	Reg#(Mantissa54) b1 <- mkRegU;
	Reg#(Index64) mso2 <- mkRegU;
       	Reg#(Mantissa54) sum2 <- mkRegU;
	Reg#(Mantissa54) sum3 <- mkRegU;
       	Reg#(Sign) sign <- mkRegU;
	Reg#(Sign) sign2 <- mkRegU;
	Reg#(Sign) sign3 <- mkRegU;
       	Reg#(Exponent) exp <- mkRegU;
	Reg#(Exponent) exp2 <- mkRegU;
	Reg#(Exp_diff) exp3 <- mkRegU;
        Reg#(Bit#(1)) op <- mkRegU;
	Reg#(Bit#(1)) op2 <- mkRegU;
	Reg#(Bit#(2)) specialOutput <- mkRegU;
	Reg#(Bit#(2)) specialOutput2 <- mkRegU;
	Reg#(Bit#(2)) specialOutput3 <- mkRegU;
	Reg#(Bool) conditionA <- mkRegU;
	Reg#(Bool) conditionB <- mkRegU;
	Reg#(Bool) conditionC <- mkRegU;
	Reg#(Bool) conditionD <- mkRegU;
	Reg#(Bool) conditionA2 <- mkRegU;
	Reg#(Bool) conditionB2 <- mkRegU;
	Reg#(Bool) conditionE2 <- mkRegU;
	Reg#(Bit#(1)) extraBit <- mkRegU;
	Reg#(Mantissa54) sum_special_extra_bit <- mkRegU;
	Reg#(Bool) dontRound <- mkRegU;
	Reg#(Bool) noOverlap <- mkRegU;
	Reg#(Bool) noOverlap3 <- mkRegU;
	Reg#(Bool) exp3isNegative <- mkRegU;
	Reg#(Bool) roundTrue <- mkRegU;
	Reg#(Bool) datap1 <- mkReg(False);
	Reg#(Bool) datap2 <- mkReg(False);
	Reg#(Bool) datap3 <- mkReg(False);

	//cycle34
	Reg#(Mantissa) mantShiftedSum_34 <- mkRegU;
	Reg#(Bit#(2)) specialOutput3_34 <- mkRegU;
	Reg#(Bool) exp3isNegative_34 <- mkRegU;
	Reg#(Mantissa54) sum3_34 <- mkRegU;
	Reg#(Sign) sign3_34 <- mkRegU;
	Reg#(Bit#(12)) expFinal11_34 <- mkRegU;
	Reg#(Mantissa) mantFinal_34 <- mkRegU;
	Reg#(Bool) noOverlap3_34 <- mkRegU;
	Reg#(Bool) datap3_34 <- mkRegU;

	//Combinational functions
	Mantissa54 sumAB = a1+b1;
        Index64 msoAB = find_mso(sumAB); 

	//cycle0 register input
	Reg#(IEEE754) a <- mkRegU;
	Reg#(IEEE754) b <- mkRegU;
	Reg#(Bool) datapi <- mkRegU;
	Reg#(Bool) datapoN <- mkRegU;

	//cycle e
	Reg#(Bool) b_bigger_than_a_reg <- mkReg(False);
	Reg#(Exp_diff) shiftedOutAmount_reg <- mkRegU;
	Reg#(IEEE754) ae <- mkRegU;
	Reg#(IEEE754) be <- mkRegU;
	Reg#(Bool) datapie <- mkRegU;

	//cycle e2
	Reg#(IEEE754) ae2 <- mkRegU;
	Reg#(IEEE754) be2 <- mkRegU;
	Reg#(Bool) datapie2 <- mkRegU;
	Reg#(Bool) b_bigger_than_a_reg2 <- mkRegU;
	Reg#(Exp_diff) shiftedOutAmount_reg2 <- mkRegU;
	Reg#(Mantissa) chooseAorB_reg <- mkRegU;
	Reg#(Mantissa) chooseAorBother_reg <- mkRegU;
	Reg#(Mantissa53) fullShiftedOutBits_reg <- mkRegU;
	Reg#(Mantissa54) fullShiftedOutBits1_reg <- mkRegU;
	Reg#(Bit#(55))fullShiftedOutBits2_reg <- mkRegU;
	Reg#(Mantissa51) shiftedOutBits_reg <- mkRegU;


	method ActionValue#(Sonly) do_sum(IEEE754 a_in, IEEE754 b_in, Bool datapi_in);

		//Combinational functions
		//cycle1 (to compute conditions)
		Exp_diff expDiff = f_exp_diff((or_exponent(a_in.e)==0)?1:a_in.e, (or_exponent(b_in.e)==0)?1:b_in.e);
		Bool b_bigger_than_a = ((expDiff<0) || ((expDiff==0) && (b_in.m>a_in.m)));
		Exp_diff shiftedOutAmount = (b_bigger_than_a)?(-expDiff):(expDiff);
		Mantissa chooseAorB = ((b_bigger_than_a_reg)?(ae2.m):(be2.m));
		Mantissa chooseAorBother = ((b_bigger_than_a_reg)?(be2.m):(ae2.m));
		Mantissa53 fullShiftedOutBits = {((b_bigger_than_a_reg)?or_exponent(ae2.e):or_exponent(be2.e)), chooseAorB};
		Mantissa54 fullShiftedOutBits1 = {fullShiftedOutBits, 1'b0};
		Bit#(55) fullShiftedOutBits2 = {fullShiftedOutBits, 2'b0};
		Mantissa51 shiftedOutBits = fullShiftedOutBits[50:0] & (~shift_left51m2(51'h7ffffffffffff, unpack(pack(shiftedOutAmount_reg))));
		Mantissa54 shiftedMantissa = ({1'b0, fullShiftedOutBits_reg} >> shiftedOutAmount_reg2);
		Bit#(1) shiftedmsb1 = fullShiftedOutBits1_reg[shiftedOutAmount_reg2];
		Bit#(1) shiftedmsb2 = fullShiftedOutBits2_reg[shiftedOutAmount_reg2];
		Bool checkNoOverlap = (shiftedOutAmount_reg2>53);
		//Compute conditions for next cycle to improve delay and have fewer registers // note, these are now computed one cycle before
		Bool conditionA2_comb = (conditionA || (sumAB[1]==1)) && (sumAB[0]==1);
		Bool conditionB2_comb = (conditionB || (sumAB[0]==1)) && (extraBit==1) && (!dontRound);
		Bool conditionE2_comb = ((conditionD || (extraBit==1)) && (conditionC));
		//Dynamic shift in cycle 3
		Exp_diff signedExp = unpack({1'b0, pack(exp2)}-{6'b000000, pack(mso2)}+1);
		//cycle4 (constants)
		Mantissa zero_mant = 0;
		Mantissa allones_mant = 52'hfffffffffffff;
		Exponent zero_exp = 0;
		Exponent allones_exp = 11'b11111111111;
		Int#(2) caseFinal = (sum3[53]==1)?1:((sum3[52]==1)?0:(-1));
		Exp_diff expFinal = exp3+((caseFinal==1)?1:((caseFinal==0)?0:(-1)));
		Bit#(12) expFinal11 = unpack(pack(expFinal));
		Mantissa mantFinal0 = (caseFinal==1)?sum3[52:1]:((caseFinal==0)?sum3[51:0]:{sum3[50:0], 1'b0});
		Mantissa mantFinal = (expFinal11[10:0]==(11'b11111111111))?(52'b0):mantFinal0;


		/////////////
		// cycle e //
		/////////////

		ae2 <= a_in;
		be2 <= b_in;
		datapie2 <= datapi_in;
		b_bigger_than_a_reg <= b_bigger_than_a;
		shiftedOutAmount_reg <= shiftedOutAmount;
		

		//////////////
		// cycle e2 //
		//////////////

		a <= ae2;
		b <= be2;
		datapi <= datapie2;
		b_bigger_than_a_reg2 <= b_bigger_than_a_reg;
		shiftedOutAmount_reg2 <= shiftedOutAmount_reg;
		chooseAorB_reg <= chooseAorB;
		chooseAorBother_reg <= chooseAorBother;
		fullShiftedOutBits_reg <= fullShiftedOutBits;
		fullShiftedOutBits1_reg <= fullShiftedOutBits1;
		fullShiftedOutBits2_reg <= fullShiftedOutBits2;
		shiftedOutBits_reg <= shiftedOutBits;
		

		/////////////
		// cycle 1 //
		/////////////

	        //subtract exponents, shift mantissas
		if (b_bigger_than_a_reg2) begin //B is bigger than A
		   sign <= b.s;
		   exp <= (or_exponent(b.e)==0)?1:b.e;
		end else begin //A is bigger than B
		   sign <= a.s;
		   exp <= (or_exponent(a.e)==0)?1:a.e;
		end
		op <= b.s ^ a.s;
			a1 <= {1'b0, ((b_bigger_than_a_reg2)?or_exponent(b.e):or_exponent(a.e)), chooseAorBother_reg};
			b1 <= ((b.s^a.s)==0) ? shiftedMantissa : (-shiftedMantissa);
			conditionA <= ((shiftedmsb1==1) || (shiftedmsb2==1) || or_Mantissa51(shiftedOutBits_reg));
			conditionB <= ((shiftedmsb2==1) || or_Mantissa51(shiftedOutBits_reg));
			conditionC <= (shiftedmsb2==1);
			conditionD <= or_Mantissa51(shiftedOutBits_reg);
			extraBit <= shiftedmsb1;
			dontRound <= checkNoOverlap;
		//Check if special input
		// specialOutput = 00: normal, 01: NaN, 10: inf, 11: -inf
		if(((a.e==(~11'b0))&&(a.m!=0)) || ((b.e==(~11'b0))&&(b.m!=0))) begin //case NaN
			specialOutput <= 1;
		end else if((a.e==(~11'b0))&&(a.m==0)) begin //case a is inf
			if(((b.e==(~11'b0))&&(b.m==0)) && (b.s!=a.s)) begin //case b is inf too and their signs are different, return NaN
				specialOutput <= 1;
			end else begin //else return inf
				specialOutput <= {1'b1, a.s};
			end
		end else if((b.e==(~11'b0))&&(b.m==0)) begin //case b is inf, return it
			specialOutput <= {1'b1, b.s};
		end else begin
			specialOutput <= 0;
		end
		datap1 <= datapi;


		/////////////
		// cycle 2 //
		/////////////

		//perform add and compute MSO
		sum2 <= sumAB;
		mso2 <= msoAB;
		sign2 <= sign;
		exp2 <= exp;
		op2 <= op;
		noOverlap <= dontRound;
		//Register rounding conditions
		conditionA2 <= conditionA2_comb;
		conditionB2 <= conditionB2_comb;
		conditionE2 <= conditionE2_comb;
		roundTrue <= ((conditionA2_comb && (sumAB[53]==1))||(conditionB2_comb && (sumAB[52]==1) && (sumAB[53]==0))||(conditionE2_comb && (sumAB[51]==1) && (sumAB[53]==0) && (sumAB[52]==0)));
		//Update special cases
		specialOutput2 <= specialOutput;

		sum_special_extra_bit <= {sumAB[52:0], 1'b0}-((extraBit==1)?1:0); // do this one cycle earlier to improve critical path

		datap2 <= datap1;


		/////////////
		// cycle 3 //
		/////////////

		//Re-alignment, rounding
		//Update sum mantissa
		if(mso2==0) begin
			//overflow
			//msb sum2[0], lsbs all shiftedOutBits
			// (sumAB[0]==0) --> no rounding needed, case 0-1 & 0-0
			// Else round up, case 1-1 // conditionA
			// if notA then case 1-0, tie
			sum3 <= ({1'b0, sum2[53:1]}+((conditionA2)?1:0));
		end else if (mso2==1) begin
			//standard
			//msb is msb of shiftedOutBits, lsbs are all shiftedOutBits except msb
			//conditionB
			//sum3 <= (sum2+((conditionB2)?((op2==0)?(1):(-1)):0));
			sum3 <= ((conditionB2)?((op2==0)?(sum2+1):(sum2-1)):sum2);
		end else if (mso2==2) begin
			//underflow by 1
			//msb of shiftedOutBits is lsb of mantissa. For rounding: msb is second msb of shiftedOutBits, lsbs are rest of shiftedOutBits
			//conditionC and conditionD combining into conditionE
			sum3 <= (sum_special_extra_bit - ((conditionE2)?1:0));
		//end else if ((mso2==53) &&(sum2[0]==0)) // sum is all 0s, 0 output (no need for this, handled by next case)
		end else begin
			// underflow by more than 1 (also handles case of all 0 mantissa output)
			//sum3 <= ({1'b0, sum_special_extra_bit}<<mso2)[54:1];
			//sum3 <= (sum_special_extra_bit<<(mso2-1));
			sum3 <= shift_left55({1'b0, sum_special_extra_bit}, mso2)[54:1];
			//(sum_special_extra_bit<<(mso2-1)); //shift in rounding value
		end
		//update exponents
		exp3 <= signedExp;
		exp3isNegative <= (signedExp<0);
		//Update signs
		sign3 <= sign2;
		//Update other variables
		specialOutput3 <= specialOutput2;
		noOverlap3 <= noOverlap;

		datap3 <= datap2;


		//////////////
		// cycle 34 //
		//////////////

		mantShiftedSum_34 <= ({1'b1, mantFinal[51:1]}>>(-expFinal+1));
		specialOutput3_34 <= specialOutput3;
		exp3isNegative_34 <= exp3isNegative;
		sum3_34 <= sum3;
		sign3_34 <= sign3;
		expFinal11_34 <= expFinal11;
		mantFinal_34 <= mantFinal;
		noOverlap3_34 <= noOverlap3;
		datap3_34 <= datap3;

	
		/////////////
		// cycle 4 //
		/////////////

		//Check for special cases, 1-bit alignment due to rounding, return results
		if(specialOutput3_34!=0) begin //special output known from beginning
			srs <= unpack(pack(tuple3(specialOutput3_34[0], allones_exp, ((specialOutput3_34[1]==0)?(allones_mant):(zero_mant)))));
		end else begin

			//Output sum

			if ((exp3isNegative_34) || (!or_Mantissa54(sum3_34))) begin // Denorm or 0 result from computation				
				srs <= unpack(pack(tuple3(sign3_34, zero_exp, (!or_Mantissa54(sum3_34))?zero_mant:mantShiftedSum_34)));
			end else begin
				srs <= unpack(pack(tuple3(sign3_34, expFinal11_34[10:0], mantFinal_34)));
			end

		end

		datapo <= datap3_34;

		return s;

	endmethod

endmodule

endpackage
