/************************************************************************************************
 * Copyright  (c) 2011, 2012, 2013 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

/*** Types for FPAR and TS code ***/
/***	
Author: Edin Kadric
	IC Group, University of Pennsylvania

Notes: Code based on the architecture described in 
	"Accurate Parallel Floating-Point Accumulation", in Arith 2013

***/

import Vector::*;

Integer n0 = 4096; // change these for changing N
Integer logN0m2 = 10; // change these for changing N
Bit#(12) n00 = 12'd4094; // change these for changing N // Number of nodes in tree minus 1
Bit#(11) n01 = 11'd2047; // change these for changing N
Bit#(10) n02 = 10'd1023; // change these for changing N
typedef 12 LogN; // change these for changing N
typedef 11 LogNm1; // change these for changing N
typedef 10 LogNm2; // change these for changing N

/* //for debug
Integer n0 = 16; // change these for changing N
Integer logN0m2 = 2; // change these for changing N
Bit#(4) n00 = 4'd14; // change these for changing N // Number of nodes in tree minus 1
Bit#(3) n01 = 3'd7; // change these for changing N
Bit#(2) n02 = 2'd3; // change these for changing N
typedef 4 LogN; // change these for changing N
typedef 3 LogNm1; // change these for changing N
typedef 2 LogNm2; // change these for changing N
*/
/* //for debug
Integer n0 = 32; // change these for changing N
Integer logN0m2 = 3; // change these for changing N
Bit#(5) n00 = 5'd30; // change these for changing N // Number of nodes in tree minus 1
Bit#(4) n01 = 4'd15; // change these for changing N
Bit#(3) n02 = 3'd7; // change these for changing N
typedef 5 LogN; // change these for changing N
typedef 4 LogNm1; // change these for changing N
typedef 3 LogNm2; // change these for changing N
*/

typedef Bit#(1) Sign;
typedef UInt#(11) Exponent;
typedef Int#(12) Exp_diff;
typedef Bit#(52) Mantissa;
typedef Bit#(54) Mantissa54; // 52 mantissa, 1 implicit bit, 1 overflow bit
typedef Bit#(55) Mantissa55; // Also keep one bit of overflow during shifting
typedef Bit#(53) Mantissa53; // For residue computations
typedef Bit#(51) Mantissa51;
typedef Bit#(54) String54;
typedef Bit#(64) String64;
typedef Bit#(6) Index64;
typedef Bit#(LogNm1) Addr1;
typedef Bit#(LogNm2) Addr2;

typedef struct{
	Sign s;
	Exponent e;
	Mantissa m;
} IEEE754 deriving (Bits, Eq, Bounded);

typedef struct{
	IEEE754 fpn;
	Bool datap;
} MemData deriving (Bits, Eq, Bounded);

typedef struct{
	IEEE754 s;
	IEEE754 r;
	Bool datap;
} SR deriving (Bits, Eq, Bounded);

typedef struct{
	IEEE754 s;
	Bool datap;
} Sonly deriving (Bits, Eq, Bounded);

function Exp_diff f_exp_diff(Exponent a, Exponent b);
	 return unpack({1'b0, pack(a)}-{1'b0, pack(b)});
endfunction


	function Bool or_Mantissa(Mantissa ip_string);     

		return  
((ip_string[0] | ip_string[1] | ip_string[2] | ip_string[3] | ip_string[4] | ip_string[5] | ip_string[6] | ip_string[7] |
ip_string[8] | ip_string[9] | ip_string[10] | ip_string[11] | ip_string[12] | ip_string[13] | ip_string[14] | ip_string[15] |
ip_string[16] | ip_string[17] | ip_string[18] | ip_string[19] | ip_string[20] | ip_string[21] | ip_string[22] | ip_string[23] |
ip_string[24] | ip_string[25] | ip_string[26] | ip_string[27] | ip_string[28] | ip_string[29] | ip_string[30] | ip_string[31] |
ip_string[32] | ip_string[33] | ip_string[34] | ip_string[35] | ip_string[36] | ip_string[37] | ip_string[38] | ip_string[39] |
ip_string[40] | ip_string[41] | ip_string[42] | ip_string[43] | ip_string[44] | ip_string[45] | ip_string[46] | ip_string[47] |
ip_string[48] | ip_string[49] | ip_string[50] | ip_string[51])==1);

endfunction


	function Bool or_Mantissa54(Mantissa54 ip_string);     

		return  
((ip_string[0] | ip_string[1] | ip_string[2] | ip_string[3] | ip_string[4] | ip_string[5] | ip_string[6] | ip_string[7] |
ip_string[8] | ip_string[9] | ip_string[10] | ip_string[11] | ip_string[12] | ip_string[13] | ip_string[14] | ip_string[15] |
ip_string[16] | ip_string[17] | ip_string[18] | ip_string[19] | ip_string[20] | ip_string[21] | ip_string[22] | ip_string[23] |
ip_string[24] | ip_string[25] | ip_string[26] | ip_string[27] | ip_string[28] | ip_string[29] | ip_string[30] | ip_string[31] |
ip_string[32] | ip_string[33] | ip_string[34] | ip_string[35] | ip_string[36] | ip_string[37] | ip_string[38] | ip_string[39] |
ip_string[40] | ip_string[41] | ip_string[42] | ip_string[43] | ip_string[44] | ip_string[45] | ip_string[46] | ip_string[47] |
ip_string[48] | ip_string[49] | ip_string[50] | ip_string[51]| ip_string[52]| ip_string[53])==1);

endfunction

	function Bool or_Mantissa53(Mantissa53 ip_string);     

		return  
((ip_string[0] | ip_string[1] | ip_string[2] | ip_string[3] | ip_string[4] | ip_string[5] | ip_string[6] | ip_string[7] |
ip_string[8] | ip_string[9] | ip_string[10] | ip_string[11] | ip_string[12] | ip_string[13] | ip_string[14] | ip_string[15] |
ip_string[16] | ip_string[17] | ip_string[18] | ip_string[19] | ip_string[20] | ip_string[21] | ip_string[22] | ip_string[23] |
ip_string[24] | ip_string[25] | ip_string[26] | ip_string[27] | ip_string[28] | ip_string[29] | ip_string[30] | ip_string[31] |
ip_string[32] | ip_string[33] | ip_string[34] | ip_string[35] | ip_string[36] | ip_string[37] | ip_string[38] | ip_string[39] |
ip_string[40] | ip_string[41] | ip_string[42] | ip_string[43] | ip_string[44] | ip_string[45] | ip_string[46] | ip_string[47] |
ip_string[48] | ip_string[49] | ip_string[50] | ip_string[51]| ip_string[52])==1);

endfunction

	function Bool or_Mantissa50(Bit#(50) ip_string);     

		return  
((ip_string[0] | ip_string[1] | ip_string[2] | ip_string[3] | ip_string[4] | ip_string[5] | ip_string[6] | ip_string[7] |
ip_string[8] | ip_string[9] | ip_string[10] | ip_string[11] | ip_string[12] | ip_string[13] | ip_string[14] | ip_string[15] |
ip_string[16] | ip_string[17] | ip_string[18] | ip_string[19] | ip_string[20] | ip_string[21] | ip_string[22] | ip_string[23] |
ip_string[24] | ip_string[25] | ip_string[26] | ip_string[27] | ip_string[28] | ip_string[29] | ip_string[30] | ip_string[31] |
ip_string[32] | ip_string[33] | ip_string[34] | ip_string[35] | ip_string[36] | ip_string[37] | ip_string[38] | ip_string[39] |
ip_string[40] | ip_string[41] | ip_string[42] | ip_string[43] | ip_string[44] | ip_string[45] | ip_string[46] | ip_string[47] |
ip_string[48] | ip_string[49])==1);

endfunction

	function Bool or_Mantissa51(Mantissa51 ip_string);     

		return  
((ip_string[0] | ip_string[1] | ip_string[2] | ip_string[3] | ip_string[4] | ip_string[5] | ip_string[6] | ip_string[7] |
ip_string[8] | ip_string[9] | ip_string[10] | ip_string[11] | ip_string[12] | ip_string[13] | ip_string[14] | ip_string[15] |
ip_string[16] | ip_string[17] | ip_string[18] | ip_string[19] | ip_string[20] | ip_string[21] | ip_string[22] | ip_string[23] |
ip_string[24] | ip_string[25] | ip_string[26] | ip_string[27] | ip_string[28] | ip_string[29] | ip_string[30] | ip_string[31] |
ip_string[32] | ip_string[33] | ip_string[34] | ip_string[35] | ip_string[36] | ip_string[37] | ip_string[38] | ip_string[39] |
ip_string[40] | ip_string[41] | ip_string[42] | ip_string[43] | ip_string[44] | ip_string[45] | ip_string[46] | ip_string[47] |
ip_string[48] | ip_string[49] | ip_string[50])==1);

endfunction

	function Bit#(1) or_exponent(Exponent ip_string);     

		return  
(ip_string==0)?(1'b0):(1'b1);


endfunction


	function Mantissa55 shift_left55(Mantissa55 ip_string, Bit#(6) amount);     

		Mantissa55 thirthysecond = ((amount[0]==1)?({ip_string[53:0], 1'b0}):(ip_string));
		Mantissa55 sixteenth = ((amount[1]==1)?({thirthysecond[52:0], 2'b0}):(thirthysecond));
		Mantissa55 eighth = ((amount[2]==1)?({sixteenth[50:0], 4'b0}):(sixteenth));
		Mantissa55 quarter = ((amount[3]==1)?({eighth[46:0], 8'b0}):(eighth));
		Mantissa55 half = ((amount[4]==1)?({quarter[38:0], 16'b0}):(quarter));
		Mantissa55 dynamicallyShiftedMantissa = ((amount[5]==1)?({half[22:0], 32'b0}):(half));
		return dynamicallyShiftedMantissa;

	endfunction

	// << (amount-2), where amount is 12-bit
	function Mantissa51 shift_left51m2(Mantissa51 ip_string, Bit#(12) amount);     

		Mantissa53 thirthysecond = ((amount[0]==1)?({1'b0, ip_string, 1'b0}):({2'b0, ip_string}));
		Mantissa53 sixteenth = ((amount[1]==1)?({thirthysecond[50:0], 2'b0}):(thirthysecond));
		Mantissa53 eighth = ((amount[2]==1)?({sixteenth[48:0], 4'b0}):(sixteenth));
		Mantissa53 quarter = ((amount[3]==1)?({eighth[44:0], 8'b0}):(eighth));
		Mantissa53 half = ((amount[4]==1)?({quarter[36:0], 16'b0}):(quarter));
		Mantissa53 dynamicallyShiftedMantissa = ((amount[5]==1)?({half[20:0], 32'b0}):(half));
		Mantissa51 rightShiftedBy2 = (((amount[6]|amount[7]|amount[8]|amount[9]|amount[10])==1)?(0):(dynamicallyShiftedMantissa[52:2]));
		return rightShiftedBy2;

	endfunction

	// << (amount-2), where amount is 12-bit
	function Mantissa51 shift_left51m2_v2(Mantissa51 ip_string, Bit#(12) amount);     

		Mantissa53 thirthysecond = ((amount[0]==1)?({1'b0, ip_string, 1'b0}):({2'b0, ip_string}));
		Mantissa53 sixteenth = ((amount[1]==1)?({thirthysecond[50:0], 2'b0}):(thirthysecond));
		Mantissa53 eighth = ((amount[2]==1)?({sixteenth[48:0], 4'b0}):(sixteenth));
		Mantissa53 quarter = ((amount[3]==1)?({eighth[44:0], 8'b0}):(eighth));
		Mantissa53 half = ((amount[4]==1)?({quarter[36:0], 16'b0}):(quarter));
		Mantissa53 dynamicallyShiftedMantissa = {2'b0, ip_string}<<amount;
		Mantissa51 rightShiftedBy2 = dynamicallyShiftedMantissa[52:2];
		return rightShiftedBy2;

	endfunction

	// >> (amount), where amount is 12-bit
	function Mantissa54 shift_right54(Mantissa54 ip_string, Bit#(12) amount);     

		Mantissa54 thirthysecond = ((amount[0]==1)?({1'b0, ip_string[53:1]}):(ip_string));
		Mantissa54 sixteenth = ((amount[1]==1)?({2'b0, thirthysecond[53:2]}):(thirthysecond));
		Mantissa54 eighth = ((amount[2]==1)?({4'b0, sixteenth[53:4]}):(sixteenth));
		Mantissa54 quarter = ((amount[3]==1)?({8'b0, eighth[53:8]}):(eighth));
		Mantissa54 half = ((amount[4]==1)?({16'b0, quarter[53:16]}):(quarter));
		Mantissa54 dynamicallyShiftedMantissa = ((amount[5]==1)?({32'b0, half[53:32]}):(half));
		Mantissa54 withLastBitsEffect = (((amount[6]|amount[7]|amount[8]|amount[9]|amount[10])==1)?(0):(dynamicallyShiftedMantissa));
		return withLastBitsEffect;

	endfunction

	// >> (amount), where amount is 12-bit
	function Mantissa shift_right52(Mantissa ip_string, Bit#(12) amount);     

		Mantissa thirthysecond = ((amount[0]==1)?({1'b0, ip_string[51:1]}):(ip_string));
		Mantissa sixteenth = ((amount[1]==1)?({2'b0, thirthysecond[51:2]}):(thirthysecond));
		Mantissa eighth = ((amount[2]==1)?({4'b0, sixteenth[51:4]}):(sixteenth));
		Mantissa quarter = ((amount[3]==1)?({8'b0, eighth[51:8]}):(eighth));
		Mantissa half = ((amount[4]==1)?({16'b0, quarter[51:16]}):(quarter));
		Mantissa dynamicallyShiftedMantissa = ((amount[5]==1)?({32'b0, half[51:32]}):(half));
		Mantissa withLastBitsEffect = (((amount[6]|amount[7]|amount[8]|amount[9]|amount[10])==1)?(0):(dynamicallyShiftedMantissa));
		return withLastBitsEffect;

	endfunction

	// >> (amount+1)
	function Mantissa shift_right52p1(Mantissa ip_string, Bit#(12) amount);     

		Mantissa withLastBitsEffect = ip_string>>amount;
		return {1'b0, withLastBitsEffect[51:1]};

	endfunction


